/* $Id: mapchan.c,v 1.4 2003/10/25 07:21:42 yuk Exp $ */

#include <stdio.h>
#include <ctype.h>
#include <unistd.h>
#include <fcntl.h>
#include <sys/time.h>
#include <sys/ioctl.h>
#include <sys/types.h>
#include <sys/wait.h>
#include <errno.h>
#include <termios.h>
#include <string.h>
#include <stdlib.h>
#include <signal.h>

#ifndef MAX_SHELL_ARGS
  #define MAX_SHELL_ARGS 16
#endif

#ifndef CPATH
  #define CPATH "/etc/mapchan"
#endif

#define BUFSIZE 4096

#define dflt_shell "/bin/sh"

#define inputF 1
#define outputF 2
#define deadF  4
#define composeF 8 
#define bellF 16
#define controlF 32
 
#define kw_num 6
char* kw[kw_num] = { "input", "output", "dead", "compose", "bell", "control" } ;
unsigned char deadC = '\0';
unsigned char composeC = '\0';
unsigned char mapon = '\0';
unsigned char mapoff = '\0';
int was_dead = 0;
unsigned char **outtbl = NULL;
unsigned char **intbl = NULL;
unsigned char **deadtbl = NULL;


/*
         mapchan -q -s shell -f file -h -?
 */

char usage[] = { "\nUsage:\n\nmapchan [-q] [-s shell] [-f file]\
 [-h]\n\nwhere:\n\
\tshell     - shell, if not /bin/sh
\tfile      - input file\n\
\t-q	    - force and quiet
\t-h    - print this screen and exit.\n" };

int pty_fd;     /*  master */
int tty_fd;     /*  slave */
pid_t pid;	   /* child's PID      */

fd_set readfds; /* select flags     */

char *shell_prog;

int quiet = 0;
int fast_quit = 0;

extern int optind, opterr, optopt;
extern char *optarg;

help() {
  if (!quiet) printf("\r\nmapchan started\r\n");
}  

#ifdef NOBASENAME
char *basename(char *path) {
  char *p;
  if (p=strrchr(path, '/')) p++;
  else p=path;
  return(p);
}
#endif

/*
 *  
 */
int
deinit_mscr()
{

#ifdef DEBUG
   fprintf(stderr, "Closing pty\n");
#endif   
   kill(pid, SIGHUP);
   close(tty_fd);
   close(pty_fd);
}

create_child()
{
   pid_t n;
   char *p;
   char *shell_args[MAX_SHELL_ARGS];
   int i;
   
   p=shell_prog; 
   i=0; 
   while (*p) {
      while (isspace(*p)) if (*p=='\0') break;
                          else p++;
      if (*p!='\0') {
         shell_args[i++] = p;
         if (i > MAX_SHELL_ARGS - 1) {
            fprintf(stderr, "mapchan: too many args for shell\n");
            exit(1);
         }
         while (!isspace(*p)) if (*p=='\0') break;
                              else p++;
         if (*p!='\0') *(p++) = '\0';
      }
   }   
   shell_args[i] = NULL;                          

   n = fork();
   if (n < 0) {
       perror("fork");
       exit(1);
   }
   if (!n) {
       /* child */
       close(pty_fd);

       /* Set up the slave pty as a controlling terminal. */
       setsid();

       /* Create stdin, stdout, and stderr. */
       dup2(tty_fd, 0);
       dup2(tty_fd, 1);
       dup2(tty_fd, 2);
       if (tty_fd > 2)
               close(tty_fd);
               

       /* Now pass the file descriptors to the shell. */
       if (execvp(shell_prog, shell_args) == -1) {
           perror(shell_prog);
           exit(1);
       }
/*
       if (execl(shell_prog, (char*)basename(shell_prog), NULL) == -1) {
           perror(shell_prog);
           exit(1);
       }
*/
   }

   /* parent */
   pid = n;
   close(tty_fd);

   FD_ZERO(&readfds);
   FD_SET(0, &readfds);
   FD_SET(pty_fd, &readfds);
}

void
sigchld(sig)
int sig;
{
  pid_t pid;
  int status;
  int i;
  
  pid = waitpid(WAIT_ANY, &status, WNOHANG);
#ifdef DEBUG  
  fprintf(stderr, "\nChild %d died with status %d \n", pid, status);
#endif
  fast_quit = 1;

  signal(SIGCHLD, sigchld);
}

void
sighup(sig)
{
  int i;

#ifdef DEBUG
fprintf(stderr, "HUP signal detected: %d\n", sig);
#endif

  for (i=1; i<NSIG; i++) signal(i, SIG_DFL);
  fast_quit = 1;
  
}  

/*
 *       Esc-
 *  

      :
   0777 -  
   0xAA -  
   99 -  
   'A'  -  A
   A    -  A
 */

char *parse_cod      /*      NULL */
(unsigned char *ks,  /*                         */
unsigned char *out)  /*     */
{
   int i, c;
   
l_start:   
   i = ks[0];
   
   if ((i >= '1') && (i <= '9'))
      /*   */
      *out=atoi(ks);
   else   
     switch (i) {
     case '0':
                 if (ks[1] == 'x') 
                    /*   */
                    sscanf(&ks[2], "%x", out);
                 else
                    /*   */
                    sscanf(&ks[1], "%o", out);
                 break;
     case '\'':  if (ks[1] == '\\') {
                    ks = &ks[2];
                    goto l_start;
                 }
     		 else
     		    *out = ks[1];
                 break;
     case '#':   
                 *out = 0;
                 return(NULL);
                 break;
     default:    *out = i;
                 break;
     }
     
   i = 0;
   if (ks[0] != '\'') 
      while (!isspace(ks[i])) if (!ks[i]) return(NULL);
                              else i++;
   else {
      i++;
      while (ks[i] != '\'') if (!ks[i]) return(NULL);
                            else i++;
      i++;
   }
   while (isspace(ks[i])) if (!ks[i]) return(NULL);
                          else i++;
   
   if (!ks[i]) return(NULL);
   
   return(&ks[i]);
}

keyword(char *s) {
  int i;
  
  for (i=0; i<kw_num; i++) if (!strncmp(kw[i], s, strlen(kw[i]))) break;
  if (i == kw_num) i = 0;
  else i = 1 << i;
  return(i);
}

/*
 *   
 *
 */
int load_map(fn)
char *fn;       /*      */
{
  register int i, j;
  char bufin[BUFSIZE], buf[BUFSIZE];
  int k;
  int flag;
  FILE *mf;
  char *p;
  unsigned char c;

  flag = 0;
  
  for(i=0;isgraph(fn[i]); i++);
  fn[i] = '\0';
  if ((mf=fopen(fn, "r")) == NULL) {       /*    */
     /*     -   CPATH */
     strncpy(buf, CPATH, BUFSIZE-2);
     strcat(buf, "/");
     strncat(buf, fn, BUFSIZE-strlen(buf)-2);
     if ((mf=fopen(buf, "r")) == NULL) {
        /*     -   */
        perror(buf);
        return(1);
     }
  }
  /*   .       */
  j = 0;
  while (fgets(bufin, 100, mf) != NULL) {
     j++;
     if (isspace(bufin[0]) || (bufin[0] == '#') || (!strlen(bufin)))
        continue;
     if (i=keyword(bufin))
        switch (i) {
        case inputF:
                    	flag = i;
                    	intbl = calloc(sizeof(char*), 256);
                    	break;
        case outputF:
                    	flag = i;
                    	outtbl = calloc(sizeof(char*), 256);
                    	break;
        case controlF:  
        		flag = i;
        		break;
        case bellF:
        case composeF:
                        break;
        case deadF:
        		if (flag & inputF) flag |= i;
        		else {
          		   fprintf(stderr, "error in file %s, str. %d\n",
        		           fn, j);
        		   return(1);
        		}
        		for (i=0; i<strlen(bufin); i++) 
        		  if (p=parse_cod(bufin, &c))
        		     parse_cod(p, &deadC);
                    	deadtbl = calloc(sizeof(char*), 256);
        		break;
        }
     else {
        p = parse_cod(bufin, &c);
        for (i=0; p != NULL; i++) {
          if (i>=99) {
             buf[i] = '\0';
             break;
          }
          p=parse_cod(p, &buf[i]);
        }
        if (flag == outputF) {
           outtbl[c] = malloc(i);
           memcpy(outtbl[c], buf, i);
           outtbl[c][i] = '\0';
        }
        else if (flag & inputF) {
                if (flag & deadF) {
                   deadtbl[c] = malloc(i);
                   memcpy(deadtbl[c], buf, i);
                   deadtbl[c][i] = '\0';
                }
                else {
                   intbl[c] = malloc(i);
                   memcpy(intbl[c], buf, i);
                   intbl[c][i] = '\0';
                }
             }
        else if (flag == controlF) {
           switch (c) {
           case '-':	/* mapping off */
                mapoff = buf[0];
                break;
           case '+':	/* mapping on */
                mapon = buf[0];
                break;
           default:
           	break;
           }
        }
     }        
  }      
  fclose(mf);
  return(0);
}        


char *default_tbl() {
  int i;
  char buf[BUFSIZE], buf1[BUFSIZE];
  char *p, *r;
  FILE *f;
  
  if (p=(char*)basename(ttyname(fileno(stdin)))) {
     strncpy(buf, CPATH, BUFSIZE-1);
     strncat(buf, "/default", BUFSIZE-strlen(buf));
     if (f=fopen(buf, "r")) {
        while (fgets(buf, BUFSIZE, f)) {
          if (!isgraph(buf[0]) || (buf[0] == '#')) 
             continue;
          for (i=0; isgraph(buf[i]); i++) buf1[i] = buf[i]; 
          buf1[i] = '\0';
          if (!strcmp(p, buf1)) {
             for (i=0; isgraph(buf[i]); i++);
             for (; isspace(buf[i]); i++);
             r = malloc(strlen(buf)+1);
             strcpy(r, &buf[i]);
             break;
          }
          else r = NULL;
        }
        fclose(f);
     }
     else {
       perror(buf);
       r=NULL;
     }
  }
  return(r);
}  
          
#define flush_in() while (j > 0) { \
                                m = write(pty_fd, p, j); \
                                if (m < 0) { \
                                        perror("write pty_fd"); \
                                        break; \
                                } \
                                p += m; j -= m; \
                              }

#define flush_out() while (j > 0) { \
                                   m = write(1, p, j); \
                                   if (m < 0) { \
                                        perror("write 1"); \
                                        break; \
                                   } \
                                   p += m; j -= m; \
                                  }


main(argc, argv)
int argc;
char **argv;

{
        int i, m, n, j, k;
        unsigned char buf[BUFSIZE], buf1[BUFSIZE], c, *p, **pd;
        fd_set lreadfds; /* select flags     */
        struct termios tios, old_tios;
        int errflg;
        char *tbl_file;
        int domapping = 1;

        /*
         *    
         */

        /* Defaults */
        shell_prog = dflt_shell;
        tbl_file = NULL;

        /*      */
        errflg = 0;
        while ((i = getopt(argc, argv, "qs:f:h?")) != EOF) {
          switch (i) {

	  case 's': /*  shell  */
		    shell_prog = optarg;
		    break;
          case 'f': /*   */
                    tbl_file = optarg;
          	    break;
          case 'q':
          	    quiet = 1;
          	    break;
          case 'h':
          case '?':
                    errflg++;
                    break;
                    
          default:  /*    -                    */
                    fprintf(stderr, "Unknown key: -%c\n", i);
                    errflg++;
                    break;
          }
        }
        if (errflg) {                 /*       */
           fprintf(stderr, usage);    /*               */
           exit(1);                   /*                */
        }
        
        if (!tbl_file) tbl_file = default_tbl();
#ifdef DEBUG
	fprintf(stderr, "Reading file %s\n", tbl_file);
#endif        
           
	if ((tbl_file == NULL) || load_map(tbl_file) ||
	   ((outtbl == NULL) && (intbl == NULL) && (deadtbl == NULL))) {
	   if (quiet) {
              if (execl(shell_prog, (char*)basename(shell_prog), NULL) == -1) {
                  perror(shell_prog);
                  exit(1);
              }
	   }
	   else {
 	      fprintf(stderr, "You don't define no one table - mapchan not needed\n");
              fprintf(stderr, usage);    /*               */
	      return(1);
	   }
	}   

#ifdef DEBUG
 if (outtbl) {
  printf("output\n");
  for (i=0; i<256; i++) if (outtbl[i]) { 
                           printf("   %c=%#0x) ", i, i);
                           for (j=0; j<strlen(outtbl[i]); j++)
                             if (outtbl[i][j] <= '\040') 
                                printf("%o ", outtbl[i][j] & 0377);
                             else
                                printf("%c ", outtbl[i][j]);
                           putchar('\n');
                         }
 } 
 if (intbl) {
  printf("input\n");
  for (i=0; i<256; i++) if (intbl[i]) { 
                           printf("   %c=%#0x) ", i, i);
                           for (j=0; j<strlen(intbl[i]); j++)
                             if (intbl[i][j] <= '\040') 
                                printf("%o ", intbl[i][j] & 0377);
                             else
                                printf("%c ", intbl[i][j]);
                           putchar('\n');
                         }
 } 
 if (deadtbl) {
  printf("dead\n");
  for (i=0; i<256; i++) if (deadtbl[i]) { 
                           printf("   %c=%#0x) ", i, i);
                           for (j=0; j<strlen(deadtbl[i]); j++)
                             if (deadtbl[i][j] <= '\040') 
                                printf("%o ", deadtbl[i][j] & 0377);
                             else
                                printf("%c ", deadtbl[i][j]);
                           putchar('\n');
                         }
 } 
#endif        
        tcgetattr(0, &old_tios);
        memcpy(&tios, &old_tios, sizeof(struct termios));
//        tios.c_iflag = 
	tios.c_iflag &= ~(ICRNL);
        tios.c_oflag = tios.c_lflag = 0;
        tios.c_cflag |= CS8 ;
        tios.c_cflag &= ~( PARENB | PARODD | CSTOPB );
          
        tcsetattr(0, TCSANOW, &tios);
        sleep(1);
        putchar(' ');
        sleep(1);
        
        signal(SIGCHLD, sigchld);
        for (i=1; i<NSIG; i++) if (i != SIGCHLD) signal(i, sighup);
        
	if (openpty(&pty_fd,&tty_fd,NULL,NULL,NULL)) {
		 perror("openpty failed: "); exit(1); };

        /* c  */
        create_child();
        
        i = 0;
        help();
        
        while (!fast_quit) {
                lreadfds = readfds;
                j = 0;
                while (select(32, &lreadfds, (fd_set *) 0, 
                          (fd_set *) 0, (struct timeval *) 0) < 0) {
                   sleep(1);
                   if (j++ > 10) {
                     perror("select");
                     break;
                   }
                }
                if (j>10) {
                 break;
                }
                if (FD_ISSET(0, &lreadfds)) {
                        n = read(0, buf, sizeof buf);
                        if (n < 0) {
                                perror("read 0");
                                break;
                        }
                        else if (intbl) {
                               p = buf1;
                               for (i=0,j=0;i<n;i++) {
                                 c = buf[i];
                                 if (deadtbl && (!was_dead) && (c == deadC)) {
                                  was_dead = 1;
                                  continue;
                                 }
                                 if (was_dead) {
                                    pd = deadtbl;
                                    was_dead = 0;
                                 }
                                 else pd = intbl;
                                 if (pd[c]) {
                                  if (j+strlen(pd[c]) >= BUFSIZE) {
                                     /****   ***/
                                     flush_in();
                                  }
                                  k = 0;
                                  while (pd[c][k]) buf1[j++] = pd[c][k++];
                                 }
                                 else {
                                  if (j+1 >= BUFSIZE) {
                                     /****   ***/
                                     flush_in();
                                  }
                                  buf1[j++] = buf[i];
                                 }
                               }
                             }  
                             else {
                               j = n;
                               p = buf;
                             }
                             flush_in();
                }

                if (FD_ISSET(pty_fd, &lreadfds)) {
                        n = read(pty_fd, buf, sizeof buf);
                        if (!n) break;
                        if (n < 0) {
                                if (errno != EIO)
                                        perror("read pty_fd");
                                continue;/*break;*/
                        }
                        else if (outtbl) {
                               p = buf1;
                               for (i=0, j=0;i<n;i++) {
                                 c = buf[i];
                                 if (domapping) {
                                   if (mapoff && (c == mapoff)) {
                                      domapping = 0;
                                      continue;
                                   }
                                 }
                                 else {
                                   if (mapon && (c == mapon)) {
                                      domapping = 1;
                                      continue;
                                   }
                                 }

                                 if (domapping && outtbl[c]) {
                                      if (j+strlen(outtbl[c]) >= BUFSIZE) {
                                         /****   ***/
                                         flush_out();
                                      }
                                      k = 0;
                                      while (outtbl[c][k]) buf1[j++] = outtbl[c][k++];
                                 }
                                 else {
                                   if (j+1 >= BUFSIZE) {
                                      /****   ***/
                                      flush_out();
                                   }
                                   buf1[j++] = buf[i];
                                 }
                               }
                             }
                             else {
                               p = buf;
                               j = n;
                             }
                             flush_out();
                }
        }
done:

        if (tcsetattr(0, TCSANOW, &old_tios) < 0)
                perror("tcsetattr oldtios");

        if (deinit_mscr()) {
            fprintf(stderr, "Deinitialization error: screen %d\n", i);
            exit(1);
        }
        if (!quiet) printf("\r\nmapchan ended.\r\n");
        return(0);
}

